#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/functional.h>

#include <graph.hpp>
#include <string>

namespace py = pybind11;

PYBIND11_MODULE( pygraph, m ) {
    m.doc() = "PV264 simple graph demo";

    using PlainGraph = Graph< std::string, Nothing >;
    using PlainNode = typename PlainGraph::NodeT;

    py::class_< PlainGraph >( m, "PlainGraph" )
        .def( py::init<>() )
        // .def( "nodes", &PlainGraph::nodes ) - cannot be used as pybind11
        // struggles to 'decipher' the nested types
        .def( "nodes", []( PlainGraph& obj ) {
            auto pylist = py::list();
            for ( auto& node : obj.nodes() ) {
                auto pyobj = py::cast( *node, py::return_value_policy::reference );
                pylist.append( pyobj );
            }
            return pylist;
        } )
        .def( "addNode", &PlainGraph::addNode, py::return_value_policy::reference )
        .def( "walkDfs", &PlainGraph::walkDfs );

    py::class_< PlainNode >( m, "PlainNode" )
        .def( "addNeighbor", py::overload_cast< PlainNode * >( &PlainNode::addNeighbor ) )
        .def_readwrite( "data", &PlainNode::data );

    using PyGraph = Graph< py::object, py::object >;
    using PyNode = typename PyGraph::NodeT;

    py::class_< PyGraph >( m, "Graph" )
        .def( py::init<>() )
        .def( "nodes", []( PyGraph& obj ) {
            auto pylist = py::list();
            for ( auto& node : obj.nodes() ) {
                auto pyobj = py::cast( *node, py::return_value_policy::reference );
                pylist.append( pyobj );
            }
            return pylist;
        } )
        .def( "addNode", &PyGraph::addNode, py::return_value_policy::reference )
        .def( "walkDfs", &PyGraph::walkDfs );
    py::class_< PyNode >( m, "Node" )
        .def( "addNeighbor", py::overload_cast< PyNode *, const py::object& >( &PyNode::addNeighbor ) )
        .def_readwrite( "data", &PyNode::data );
}